/*

 ideserv.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdio.h>
#include <errno.h>
#ifndef WIN32		//*NIX
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#else			//WIN32
#include <windows.h>
#include "resource.h"
#endif 			//ALL
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <locale.h>
#include "getopt.h"
#include "path.h"
#include "partition.h"
#include "driver.h"
#include "usb.h"
#include "eth.h"
#include "vice.h"
#include "rs232.h"
#ifndef OSX
#include "x1541.h"
#include "pc64.h"
#endif

#include "crc8.h"
#include "ideserv.h"
#include "compat.h"
#include "normal.h"
#include "message.h"

static const struct io_ops_s *io_ops;
static FILE *lf;

static struct arguments_s arguments;
static struct buffer_s buff[16];

#ifdef WIN32
/* Prototypes */
static INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
static long WINAPI PCLinkLoop(long);
/* Global variable */
static NOTIFYICONDATA g_notifyIconData;
#endif

static const char *argp_program_version="ideservd " VERSION;

static void terminate(int UNUSED(x)) {
    if (arguments.mode != M_ETHERNET) io_ops->shutdown();
    if (lf) {
	fprintf(lf,"Terminated.\n");
	fclose(lf);
    }
    exit(0);
}

void seterror(int c,int i1) {
    const char *msg="UNKNOWN ERROR";

    switch (c) {
    case 00:msg="OK";break;
    case 01:msg="FILES SCRATCHED";break;
    case 02:msg="PARTITION SELECTED";break;
    case 23:if (i1==0) msg="READ ERROR"; else msg="CRC ERROR";i1=0;break;
    case 26:msg="WRITE PROTECT ON";break;
    case 27:msg="ACCESS DENIED";break;
    case 31:msg="UNKNOWN COMMAND";break;
    case 30:
    case 33:
    case 34:msg="SYNTAX ERROR";break;
    case 39:msg="PATH NOT FOUND";break;
    case 41:msg="FRAME ERROR";break;
    case 42:msg="CRC ERROR";break;
    case 62:msg="FILE NOT FOUND";break;
    case 63:msg="FILE EXISTS";break;
    case 70:msg="NO CHANNEL";break;
    case 72:msg="PARTITION FULL";break;
    case 73:msg="IDESERV " VERSION2 " PCLINK";break;
    case 77:msg="SELECTED PARTITION ILLEGAL";break;
    }
    buff[15].length = sprintf((char *)buff[15].data,"%02d, %s,%03d,000,000,000",c,msg,i1);
    buff[15].pointer = 0;
}

volatile int tout;
static void timeout(int UNUSED(x)) {
    fprintf(lf,"Timeout.\n");
    io_ops->clean();
    tout=1;
}

void errtochannel15(int i) {
    if (i)
        switch (errno) {
        case EISDIR:
        case ENOTDIR:
        case ENOENT:seterror(62,0);break;
        case EEXIST:seterror(63,0);break;
        case EPERM:
        case EACCES:seterror(27,0);break;
        case ENOTEMPTY:seterror(1,0);break;
        case ENOSPC:seterror(72,0);break;
        case EROFS:seterror(26,0);break;
        case EIO:seterror(23,0);break;
        default:seterror(80,errno);
        }
    else seterror(0,0);
}

static void do_chdir(const c64c_t *cmd) {
    struct dirent *ep;
    c64c_t name[17];
    const c64c_t *outname;
    DIR *dp;
    char outpath[1020];
    int found=0;
    partition_t outpart = 0;

    if (!strchr((char *)cmd, ':')) {
        if (!strcmp((char *)cmd, "_")) cmd=(c64c_t *)":_";
        else if (cmd[0] != '/') { seterror(30,0); goto vege2; }
        else cmd++;
    }

    if (!(found=resolv_path(cmd, outpath, &outname, &outpart, lf, arguments.relaxfiletypes))) {seterror(39,0);goto vege2;}

    if (!outname[0]) goto vege2;
    if (!strcmp((const char *)outname,"_") || !strcmp((const char *)outname, "..")) {
        char *o = strrchr(outpath, '/');
        if (o) *o=0; else outpath[0]=0;
        goto vege2;
    }
    if (!strcmp((const char *)outname, ".")) goto vege2;
    found = 0;
    convertfilename(outname, ',', name, NULL, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        strcpy(outpath, lname2);
        found=1;
        break;
    }
    closedir(dp);
    initdirentry(1);
vege2:
    if (found) {
        partition_set_path(outpart, outpath);
        seterror(0,0);
    } else seterror(39,0);
    fprintf(lf,"Chdir to \"%s\".\n", outpath[0]?outpath:"/");
}

static void do_mkdir(const c64c_t *cmd) {
    struct dirent *ep;
    DIR *dp;
    char outpath[1020],lname[1000];
    int found=0;
    c64c_t name[17];
    const c64c_t *outname;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.relaxfiletypes)) {seterror(39,0);goto vege2;}

    convertfilename(outname, ',', name, NULL, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strncpy(lname,ep->d_name,999);
        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        found=1;
        break;
    }
    closedir(dp);
    initdirentry(1);

    if (!found) {
        mbstate_t ps;
        int f, f2;
        memset(&ps, 0, sizeof(ps));
        for (f = f2 = 0; name[f] && f < 16; f++) {
            f2 += c64toascii(lname + f2, name[f], &ps);
        }
        lname[f2]=0;
    }

    if (outpath[0]) strcat(outpath,"/");
    strcat(outpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
        errtochannel15(mkdir(outpath,0777));
    }
vege2:
    if (!name[0]) {
        seterror(34,0);
    }
    fprintf(lf,"Mkdir \"%s\".\n",outpath[0]?outpath:"/");
}

static void do_scratch(const c64c_t *cmd, int smode) {
    char outpath[1000];
    struct dirent *ep;
    DIR *dp;
    int found=0;
    c64c_t name[17], type[4]={'*',0};
    const c64c_t *outname;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.relaxfiletypes)) {seterror(39,0);goto vege;}

    convertfilename(outname, '=', name, type, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        }
        else
            if ((dirent.attrib & A_ANY)==A_DIR) continue;

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        found=1;
        if (smode) {
            fprintf(lf,"Rmdir \"%s\"\n",lname2);
            errtochannel15(rmdir(lname2));
        } else {
            errtochannel15(unlink(lname2));
            fprintf(lf,"Rm \"%s\"\n",lname2);
        }
    }
    closedir(dp);
    initdirentry(1);
    if (!found) seterror(1,0);
vege:
    if (!found) {
        if (smode) {
            fprintf(lf,"Rmdir \"%s\"\n",outpath);
        } else {
            fprintf(lf,"Rm \"%s\"\n",outpath);
        }
    }

    if (!name[0]) {
        seterror(34,0);
    }
}

void commandchannel(const c64c_t *s) {
    if (s[0]=='C' && s[1]=='D') {
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (s[0]=='C' && s[1]==0xd0) {
        fprintf(lf,"Change to partition %d\n",s[2]);
        seterror(partition_select(s[2]),s[2]);
    } else if (s[0]=='C' && s[1]=='P') {
        partition_t part = 0;
        int i = 2;
        if (!s[2]) part = partition_get_current(); else {
            while ((s[i] ^ 0x30) < 10) part = part * 10 + (s[i++] ^ 0x30);
        }
        fprintf(lf,"Change to partition %d\n",part);
        seterror(partition_select(part), part);
    } else if (s[0]=='U' && ((s[1] & 0x0f)==9 || (s[1] & 0x0f)==10)) {
        seterror(73,0);
        fprintf(lf,"Identify device\n");
    } else if (s[0]=='I') {
        seterror(0,0);
        fprintf(lf,"Initialize\n");
    } else if (s[0]=='V') {
        seterror(0,0);
        fprintf(lf,"Validate\n");
    } else if (s[0]=='M' && s[1]=='-' && s[2]=='R') { 
        int i,j,k;
        j=s[5];if (j>50) j=50;
        k=s[3];
	for (i=0;i<j;i++) buff[15].data[i]="IDE64 CARTRIDGE "[k++ & 15];
	buff[15].data[i]=0;
        fprintf(lf,"Memory read $%04x %02x\n", s[3] | (s[4] << 8), s[5]);
    } else if (s[0]) {
        seterror(31,0);
    } else fprintf(lf,"\n");
}

static void testarg(int argc,char *argv[]) {
    int c;

    for (;;) {
        static struct option long_options[]= {
#ifndef WIN32
            {"background", no_argument, 0, 'b'},
            {"user", required_argument, 0, 'u'},
            {"group", required_argument, 0, 'g'},
            {"nice", required_argument, 0, 'n'},
#endif
            {"root", required_argument, 0, 'r'},
            {"log", required_argument, 0, 'l'},
            {"allprg", no_argument, 0, 'P'},
            {"help", no_argument, 0, '?'},
            {"usage", no_argument, 0, 1},
            {"version", no_argument, 0, 'V'},
            {"hog", no_argument, 0, 'h'},
            {"debug", no_argument, 0, 'D'},
            {"device", required_argument, 0, 'd'},
            {"mode", required_argument, 0, 'm'},
            {"lptport", required_argument, 0, 'p'},
            {"ipaddress", required_argument, 0, 'i'},
            {"network", required_argument, 0, 'N'},
            {0,0,0,0}
        };
        int option_index=0;

        c=getopt_long(argc, argv,
#ifndef WIN32
                      "m:u:g:r:l:n:P?VbhDd:p:i:N:"
#else
                      "m:r:l:P?VhDd:p:i:N:"
#endif
                      , long_options, &option_index);
        if (c==-1) {
            if (optind==argc)
                break;
            c='?';
        }

        switch (c) {
        case '?':
            message(
                   "Usage: ideservd [OPTION...]\n"
                   "IDEDOS 0.9x PCLink fileserver\n"
                   "\n"
#ifndef WIN32
                   "  -b, --background\t     Fork into background\n"
                   "  -d, --device=DEVICE\t     Device (/dev/parport0 or /dev/ttyS0)\n"
                   "  -g, --group=GROUP\t     Group under we run (nogroup)\n"
#else //WIN32
                   "  -d, --device\t\t     Serial port device (COM1)\n"
#endif
                   "  -h, --hog\t\t     CPU hog (off)\n"
                   "  -i, --ipaddress=IP\t     IP address of C64 on network\n"
                   "  -l, --log=FILE\t     Logfile (stdout)\n"
                   "  -m, --mode=MODE\t     Mode (x1541, xe1541, xm1541, xa1541,\n"
                   "\t\t\t     pc64, pc64s, rs232, rs232s usb, vice, eth)\n"
#ifndef WIN32
                   "  -n, --nice=ADJUST\t     Adjust nice level (-19)\n"
#endif
                   "  -N, --network=NUM\t     Network number (0)\n"
                   "  -p, --lptport=IOPORT\t     Printer port address (0x378)\n"
                   "  -P, --allprg\t\t     Almost everything to PRG\n"
                   "  -r, --root=DIRECTORY\t     Root directory (.)\n"
#ifndef WIN32
                   "  -u, --user=USER\t     User under we run (nobody)\n"
#endif
                   "  -?, --help\t\t     Give this help list\n"
                   "      --usage\t\t     Give a short usage message\n"
                   "  -V, --version\t\t     Print program version\n"
                   "\n"
                   "Mandatory or optional arguments to long options are also mandatory or optional\n"
                   "for any corresponding short options.\n"
                   "\n"
                   "Report bugs to <soci@c64.rulez.org>.\n"
                   );
                   exit(0);
        case 1:
            message(
                   "Usage: ideservd [-bhP?V] [-d DEVICE] [-g GROUP] [-i IP] [-l FILE] [-m MODE]\n"
                   "            [-n ADJUST] [-N NUM] [-p IOPORT] [-r DIRECTORY] [-u USER]\n"
                   "            [--background] [--device DEVICE] [--group=GROUP] [--hog]\n"
                   "            [--ipaddress IP] [--log=FILE] [--mode MODE] [--nice=ADJUST]\n"
                   "            [--network NUM] [--lptport=IOPORT] [--allprg] [--root=DIRECTORY]\n"
                   "            [--user=USER] [--help] [--usage] [--version]\n"
                  );
            exit(0);
        case 'V':
            message("%s\n", argp_program_version);
            exit(0);
#ifndef WIN32
        case 'b':arguments.background=1;break;
        case 'u':arguments.user=optarg;break;
        case 'g':arguments.group=optarg;break;
        case 'n':arguments.priority=strtol(optarg,NULL,0);break;
#endif
        case 'P':arguments.relaxfiletypes=1;break;
        case 'D':arguments.debug=1;break;
        case 'l':arguments.log=optarg;break;
        case 'r':arguments.root=optarg;break;
        case 'h':arguments.hog=1;break;
        case 'd':arguments.device=optarg;break;
        case 'm':arguments.mode_name=optarg;break;
        case 'p':arguments.lptport=strtol(optarg,NULL,0) & 0xffff;break;
        case 'i':arguments.sin_addr=optarg;break;
        case 'N':arguments.network=strtol(optarg,NULL,0) & 0xff;break;
        default:
            exit(1);
        }
    }
}

static int init(int argc, char *argv[])
{
#ifndef WIN32
    struct passwd *user2 = NULL;
    struct group *group2 = NULL;
#endif
    const char *init_dev = NULL;
    int init_int = 0;
    int lastfail = 0;

    setlocale(LC_ALL, "");

    arguments.mode=M_USB;
    arguments.lptport=DEFAULT_IOPORT;
    arguments.priority = -19;
    testarg(argc, argv);
    partition_create(1, (c64c_t *)"PARTITION 1");
    partition_select(1);

#ifdef WIN32
    lf = (arguments.log && !strcmp(arguments.log, "-")) ? stdout : fopen(arguments.log ? arguments.log : "ideservd.log", "wt");
#else
    lf = arguments.log && strcmp(arguments.log, "-") ? fopen(arguments.log, "wt") : stdout;
#endif
    if (!lf) {
        lf = stdout;
        fprintf(lf, "Cannot open logfile '%s'\n", arguments.log);
    }
    fprintf(lf, "%s\n", argp_program_version);

    fprintf(lf, "Using driver ");

    arguments.mode = M_NONE;
    if (arguments.mode_name) {
        const struct {
            const char *name;
            enum e_modes mode;
        } *i, modes[] = {
#ifdef _X1541_H
            {"x1541", M_X1541}, {"xe1541", M_XE1541}, {"xm1541", M_XM1541}, {"xa1541", M_XA1541},
#endif
#ifdef _PC64_H
            {"pc64", M_PC64}, {"pc64s", M_PC64S},
#endif
#ifdef _RS232_H
            {"rs232", M_RS232}, {"rs232s", M_RS232S},
#endif
#ifdef _USB_H
            {"usb", M_USB},
#endif
#ifdef _VICE_H
            {"vice", M_VICE},
#endif
#ifdef _ETH_H
            {"eth", M_ETHERNET},
#endif
            {NULL, M_NONE}
        };
        i = modes;
        while (i->name) {
            if (!strcasecmp(arguments.mode_name, i->name)) {
                arguments.mode = i->mode;
                break;
            }
            i++;
        }
    } else arguments.mode = M_USB;

    switch (arguments.mode) {
#ifdef _X1541_H
    case M_X1541:
    case M_XE1541:
    case M_XM1541:
    case M_XA1541:
        {
            int i;
            const char *cable = "X";
            io_ops=&x1541_io_ops;
            i=arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_XE1541) {i |= 0x40000; cable = "XE";}
            if (arguments.mode==M_XM1541) {i |= 0x80000; cable = "XM";}
            if (arguments.mode==M_XA1541) {i |= 0xc0000; cable = "XA";}
            fprintf(lf, "%s using %s1541 cable", io_ops->drivername(), cable);
            if (arguments.device) {
                fprintf(lf, " on device %s\n", arguments.device);
            } else { 
                fprintf(lf, " at port %x\n", arguments.lptport);
            }
            init_dev = arguments.device;
            init_int = i;
        }
        break;
#endif
#ifdef _PC64_H
    case M_PC64:
    case M_PC64S:
        {
            int i;
            io_ops=&pc64_io_ops;
            i = arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_PC64S) i|=0x40000;
            if (arguments.device) {
                fprintf(lf, "%s on device %s\n", io_ops->drivername(), arguments.device);
            } else { 
                fprintf(lf, "%s at port %x\n", io_ops->drivername(), arguments.lptport);
            }
            init_dev = arguments.device;
            init_int = i;
        }
        break;
#endif
#ifdef _RS232_H
    case M_RS232:
    case M_RS232S:
        io_ops=&rs232_io_ops;
        if (!arguments.device) arguments.device = DEFAULT_COMPORT;
        fprintf(lf, "%s on device %s\n", io_ops->drivername(), arguments.device);
        init_dev = arguments.device;
        init_int = (arguments.mode == M_RS232S);
        break;
#endif
#ifdef _USB_H
    case M_USB:
        io_ops=&usb_io_ops;
        init_dev = arguments.device;
        fprintf(lf, "%s for device %s\n", io_ops->drivername(), arguments.device ? arguments.device : "*");
        break;
#endif
#ifdef _VICE_H
    case M_VICE:
        io_ops=&vice_io_ops;
        init_dev = arguments.sin_addr;
        init_int = arguments.network;
        fprintf(lf, "%s to %s", io_ops->drivername(), arguments.sin_addr ? arguments.sin_addr : "127.0.0.1");
        if (arguments.network) fprintf(lf, " port %d", arguments.network);
        putc('\n', lf);
        break;
#endif
#ifdef _ETH_H
    case M_ETHERNET:
        io_ops = &eth_io_ops;
        init_dev = arguments.sin_addr;
        init_int = arguments.network;
        fprintf(lf, "%s to %s", io_ops->drivername(), arguments.sin_addr ? arguments.sin_addr : "127.0.0.1");
        if (arguments.sin_addr) fprintf(lf, " network %d", arguments.network);
        putc('\n', lf);
        break;
#endif
    default: 
        fprintf(lf, "NONE for unknown mode '%s'\n", arguments.mode_name);
        exit(1);
    }

#ifndef WIN32
    {
        pid_t pid;

        if (arguments.background) if (fork()) {fprintf(lf, "Forking into background.\n");exit(0);}
        for (;;) {
            pid=fork();
            if (pid) {
                int status;
                usleep(1000000);
                waitpid(pid, &status, 0);
                lastfail = (signed char)WEXITSTATUS(status);
            } else break;
        }
    }
#endif

    do {
        int thisfail = io_ops->initialize(init_dev, init_int, lf, lastfail);
        int change = thisfail != lastfail;
        if (!thisfail) break;
        lastfail = thisfail;
        if (change) {
#ifndef WIN32
            exit(thisfail);
#else
            break;
#endif
        }
        usleep(1000000);
    } while (1);

#ifdef WIN32
    if (chdir(arguments.root ? arguments.root : "/cygdrive")) {
        if (lastfail != 2) fprintf(lf, "Changing dir to '%s' failed: %s(%d)\n", arguments.root ? arguments.root : "/cygdrive", strerror(errno), errno);
        if (arguments.root) exit(2);
    }
#else
    nice(arguments.priority);
    if (!arguments.user || !arguments.group) {
        uid_t uid = getuid(), euid = geteuid();
        if (uid <= 0 || uid != euid) {
            if (!arguments.user) arguments.user = "nobody";
            if (!arguments.group) arguments.group = "nogroup";
        }
    }
    if (arguments.user && !(user2=getpwnam(arguments.user))) {if (lastfail != 3) fprintf(lf, "User '%s' does not exists.\n", arguments.user);exit(3);}
    if (arguments.group && !(group2=getgrnam(arguments.group))) {if (lastfail != 4) fprintf(lf, "Group '%s' does not exists.\n", arguments.group);exit(4);}
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
    if (arguments.root) {
        uid_t uid = getuid(), euid = geteuid();
        if (chdir(arguments.root)) {
            if (lastfail != 5) fprintf(lf, "Changing dir to '%s' failed: %s(%d)\n", arguments.root, strerror(errno), errno);
            exit(5);
        } else {
            if (chroot(arguments.root) || chdir("/")) {
                if (lastfail != 6) fprintf(lf, "Chrooting failed: %s(%d)\n", strerror(errno), errno);
                if (uid <= 0 || uid != euid) exit(6);
            }
        }
    }
    if ((group2 && setgid(group2->gr_gid)) || (user2 && setuid(user2->pw_uid))) {if (lastfail != 7) fprintf(lf, "Could not drop privileges: %s(%d)\n", strerror(errno), errno);exit(7);}
#endif
    fclose(stdin);
    fclose(stderr);
    if (lf != stdout) fclose(stdout);
    return 0;
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
/*  int i;
  HWND hwnd_select;

  const char *pclinkmodes[] = {
            "x1541",
            "xe1541",
            "xm1541",
            "xa1541",
            "pc64",
            "pc64s",
            "rs232",
            "rs232s",
            "usb",
            "eth" };*/

  switch(message) 
  {
//   case WM_INITDIALOG:
//     hwnd_select = GetDlgItem(hwnd, IDC_SELECT);
//     for(i=0; i<11; i++) SendMessage(hwnd_select, CB_ADDSTRING, 0, (LPCTSTR)pclinkmodes[i]);
//     SendMessage(hwnd_select, CB_SETCURSEL, 8, 0); // from 0
//     return TRUE;
//   break;

   case WM_COMMAND:
	if (LOWORD(wParam) == IDC_EXIT || LOWORD(wParam) == IDCANCEL)
	{
		PostQuitMessage(0);
		EndDialog(hwnd,TRUE);
		return TRUE;
        }

	if (LOWORD(wParam) == IDC_BUTTON)
	{
	  ShowWindow(hwnd, SW_HIDE);
	}
   break;

//   case WM_SYSCOMMAND:
//	if (LOWORD(wParam) == SC_MINIMIZE)
//	{
//	   ShowWindow(hwnd, SW_HIDE);
//	}
//   break;

   case WM_USER:
	if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN)
        {
	   ShowWindow(hwnd, SW_SHOW);
   	}
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
// int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    int b;
#ifdef WIN32
    int lastfail = init(argc, argv);
#else
    init(argc, argv);
#endif
    signal(SIGALRM, timeout);
    signal(SIGTERM, terminate);
    signal(SIGINT, terminate);
    for (b=0;b<15;b++) buff[b].mode = CM_CLOSED;
    buff[15].mode = CM_ERR;
    if (!buff[15].data) {
        buff[15].data = (unsigned char *)malloc(256);
        if (!buff[15].data) exit(1); // out of memory?
    }
    seterror(73,0);

#ifdef WIN32
    HWND hwnd;
    MSG msg;

    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);
 
// TRAY ICON

    memset(&g_notifyIconData, 0, sizeof(NOTIFYICONDATA));
    g_notifyIconData.cbSize = sizeof(NOTIFYICONDATA);
    g_notifyIconData.hWnd = hwnd;
    g_notifyIconData.uID = 0;
    g_notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    g_notifyIconData.uCallbackMessage = WM_USER; 
    g_notifyIconData.hIcon = (HICON)LoadImage( NULL, TEXT("pclink.ico"), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
    strcpy(g_notifyIconData.szTip, "IDE64 PCLink");
    Shell_NotifyIcon(NIM_ADD, &g_notifyIconData);
    
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,(LPVOID)lastfail,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    Shell_NotifyIcon(NIM_DELETE, &g_notifyIconData);
    terminate(0);
    return 0;
}

long WINAPI PCLinkLoop(long lParam)
{
    int b;
    int lastfail = (int)lParam;
#endif //WIN32

    for (;;) {
        fprintf(lf,"ready.\n");
        fflush(lf);
    start2:
        crc_clear();
        b=io_ops->wait();
        if (b<0) {
#ifdef WIN32
            usleep(1000000);
            lastfail = io_ops->reinitialize(lf, lastfail);
            b=0;
#else
            terminate(0);
#endif
        }

        switch (b) {
        case 0x00:goto start2;
        case 'N':
        case 0xCE:b=openfile(io_ops, buff, lf, &arguments, b == 0xCE);break;
        case 'G':
        case 0xC7:b=readfile(io_ops, buff, lf, &arguments, b == 0xC7);break;
        case 'S':
        case 0xD3:b=writefile(io_ops, buff, lf, &arguments, b == 0xD3);break;
        case 'D':
        case 0xC4:b=closefile(io_ops, buff, lf, &arguments, b == 0xC4);break;
        case 'I':
        case 0xC9:b=statuserror(io_ops, (char *)buff[15].data, lf, &arguments, b == 0xC9);break;
        case 'O':
        case 0xCF:b=openfile_compat(io_ops, buff, lf, &arguments, b == 0xCF);break;
        case 'R':
        case 0xD2:b=readfile_compat(io_ops, buff, lf, &arguments, b == 0xD2);break;
        case 'W':
        case 0xD7:b=writefile_compat(io_ops, buff, lf, &arguments, b == 0xD7);break;
        case 'C':
        case 0xC3:b=closefile_compat(io_ops, buff, lf, &arguments, b == 0xC3);break;
        default:fprintf(lf,"Unknown command %02X\n",b);
        }
        if (b) io_ops->clean();
    }
    return 0;
}
